#include <stdint.h>
#include "hal.h"

static void bootloader_util_reset(uint32_t start_addr)
{
  asm volatile (
    "LDR   R5, [%0]  \n\t"
    //            ; Get App initial MSP for bootloader.
    "MSR   MSP, R5  \n\t"
    //            ; Set the main stack pointer to the applications MSP.
    "LDR   R0, [%0, #0x04]  \n\t"
    //       ; Load Reset handler into R0. This will be first argument to branch instruction (BX).
    "MOVS  R4, #0xFF  \n\t"
    //             ; Load ones to R4.
    "SXTB  R4, R4  \n\t"
    //                ; Sign extend R4 to obtain 0xFFFFFFFF instead of 0xFF.
    "MRS   R5, IPSR  \n\t"
    //              ; Load IPSR to R5 to check for handler or thread mode.
    "CMP   R5, #0x00  \n\t"
    //             ; Compare, if 0 then we are in thread mode and can continue to reset handler of bootloader.
    //             ; If not zero we need to exit current ISR and jump to reset handler of bootloader.
    : 
    : "r" (start_addr)
    : "r5"
      );

    asm volatile goto (
    "BNE   %l[israbort]  \n\t"
    :
    :
    :
    : israbort
    );

  asm volatile (
    "MOV   LR, R4  \n\t"
    //                ; Clear the link register and set to ones to ensure no return, R4 = 0xFFFFFFFF.
    "BX    R0  \n\t"
    //                    ; Branch to reset handler of bootloader.
    :
    :
    : "r4"
  );


israbort:

  asm volatile (
    "MOV   R5, R4  \n\t"
    //                ; Fill with ones before jumping to reset handling. We be popped as LR when exiting ISR. Ensures no return to application.
    "MOV   R6, R0 \n\t" 
// Move address of reset handler to R6. Will be popped as PC when exiting ISR. Ensures the reset handler will be executed when exist ISR.
    "MOVS  r7, #0x21 \n\t" 
      // Move MSB reset value of xPSR to R7. Will be popped as xPSR when exiting ISR. xPSR is 0x21000000 thus MSB is 0x21.
    "REV   r7, r7 \n\t" 
      // Reverse byte order to put 0x21 as MSB.
    "PUSH  {r4-r7} \n\t" 
      // Push everything to new stack to allow interrupt handler to fetch it on exiting the ISR.

    "MOVS  R4, #0x00 \n\t" 
      // Fill with zeros before jumping to reset handling. We be popped as R0 when exiting ISR (Cleaning up of the registers).
    "MOVS  R5, #0x00 \n\t" 
      // Fill with zeros before jumping to reset handling. We be popped as R1 when exiting ISR (Cleaning up of the registers).
    "MOVS  R6, #0x00 \n\t" 
      // Fill with zeros before jumping to reset handling. We be popped as R2 when exiting ISR (Cleaning up of the registers).
    "MOVS  R7, #0x00 \n\t" 
      // Fill with zeros before jumping to reset handling. We be popped as R3 when exiting ISR (Cleaning up of the registers).
    "PUSH  {r4-r7} \n\t" 
      // Push zeros (R4-R7) to stack to prepare for exiting the interrupt routine.

    "MOVS  R0, #0xF9 \n\t" 
      // Move the execution return command into register, 0xFFFFFFF9.
    "SXTB  R0, R0 \n\t" 
      // Sign extend R0 to obtain 0xFFFFFFF9 instead of 0xF9.
    "BX    R0 \n\t" 
      // No return - Handler mode will be exited. Stack will be popped and execution will continue in reset handler initializing other application.
    ".align 4"
    : 
    : 
    : "r5"
  );

}

int main(void)
{
  // check if button is pressed and select application to 
  // run based  on it.
  set_gpio_mode(P0, BUILT_IN_BUTTON_2_PIN, GPIO_MODE_INPUT, 3);
  spin(9999999);

  if (gpio_read(P0, BUILT_IN_BUTTON_2_PIN))
  {
    bootloader_util_reset(0x62000);
  }
  else
  {
    bootloader_util_reset(0x1000);
  }
}

// Startup code
__attribute__((naked, noreturn)) void _reset(void) {
  // memset .bss to zero, and copy .data section to RAM region
  extern long _sbss, _ebss, _sdata, _edata, _sidata;
  for (long *dst = &_sbss; dst < &_ebss; dst++) *dst = 0;
  for (long *dst = &_sdata, *src = &_sidata; dst < &_edata;) *dst++ = *src++;

  main();             // Call main()
  for (;;) (void) 0;  // Infinite loop in the case main() returns
}

extern void _estack(void);  // Defined in link.ld

// 16 standard and 42 nRF-specific handlers
__attribute__((section(".vectors"))) void (*const tab[16 + 42])(void) = {
    _estack, _reset};